# -*- coding: utf-8 -*-

import sys 
import os
import numpy as np
from scipy.sparse import csc_matrix
import pandas as pd
import pyomo.environ as pyo
import matplotlib.pyplot as plt

#### read input parameters from the following Excel file located in the same folder as this .py file
#ExcelInOrig='Data FI v20.xlsx'
ExcelInOrig='Data NO v20.xlsx'

checkfiles = 0   # do we write out the model instance and optimization results to external text files?
modelfile = 'optmodel.txt' 
resultfile= 'optresults.txt'

ipopt_exe = None
## specify where the solver (ipopt) is to be found
#ipopt_exe="C:\\Users\\dasilvf1\\OneDrive - Aalto University\\Masters Thesis\\Modelling\\Python\\Ipopt-win64-md\\bin\\ipopt.exe"

# read in from the single Excel sheets (this pre-sorts our input data)

# If the input file is open -> create a temporary copy to read in from
# we test this when reading in the first sheet
try:
    inputGen=pd.read_excel(ExcelInOrig, sheet_name='General data')
    deleteTemp=False
    ExcelIn=ExcelInOrig
except:    
    print("Input file is open")
    ExcelInTemp="temp_" + ExcelInOrig
    command="xcopy \"" + ExcelInOrig + "\" " + "\"" + ExcelInTemp + "*\""
    os.system(command)
    print("created a temporary copy")
    inputGen=pd.read_excel(ExcelInTemp, sheet_name='General data')
    deleteTemp=True
    ExcelIn=ExcelInTemp
    
inputGen=pd.read_excel(ExcelIn, sheet_name='General data')
inputMeas= pd.read_excel(ExcelIn, sheet_name='Measures',skiprows=1,index_col=1)
inputNoder= pd.read_excel(ExcelIn, sheet_name='Nodes',skiprows=1)
inputNetwork= pd.read_excel(ExcelIn, sheet_name='Network',skiprows=1)
inputRain= pd.read_excel(ExcelIn, sheet_name='Design rainfall',skiprows=1)
inputShares=pd.read_excel(ExcelIn, sheet_name='Fixed Shares', skiprows=1)

# delete the temporary file again in case it was created
if deleteTemp==True:
    os.remove(ExcelInTemp)
    print("deleted temporary file")
    deleteTemp=False

# split up the read-in data and assign to parameters
#DiscR=inputGen.cell(row=3,column=2).value
#print(DiscR)
DiscR=np.array(inputGen)[0,1]  # discount rate
TTE=int(np.array(inputGen)[1,1])  # number of time periods for economics modelling
TT=int(np.array(inputGen)[2,1]) # number of time periods for water modelling
RProfsel=int(np.array(inputGen)[3,1])   # selected hyetograph
RunSimulation=int(np.array(inputGen)[4,1]) # do we run a simulation (fixed NBS shares)?
OptDischarge=int(np.array(inputGen)[5,1])   # optimize wrt. discharge? (default = economy)
LimitDischarge=np.array(inputGen)[6,1]  # limit on discharge if optimize wrt. economics
LimitCosts=np.array(inputGen)[7,1]  # limit on costs if optimize wrt. discharge
NShareLimit=int(np.array(inputGen)[8,1]) # do we have individual limits on measure areas in nodes?
Outputfile=str(np.array(inputGen)[9,1]) # name of file to write results

ICost=np.array(inputMeas)[:,1]  # investment costs for each measure
MCost=np.array(inputMeas)[:,2]   # maintenance costs for each measure   
InfProf=np.array(inputMeas)[:,3] # infiltration for each measure
Retent=np.array(inputMeas)[:,4:]    # retention profile

NAreal=np.array(inputNoder)[:,1]  # available area in node n
NRain=np.array(inputNoder)[:,2]# rain in node n (yes/no)
NBSLimit=np.array(inputNoder)[:,3] # possible area limit on all NBS
NSol=np.array(inputNoder)[:,4:]  # feasibility/max. area of measure s in node n

NShareFix=np.array(inputShares)[:,1:]

nrow=inputNetwork['From node']
ncol=inputNetwork['To node']
nshare=inputNetwork['Share']
ntime=inputNetwork['Time of concentration (time unit)']

RProfAll=np.array(inputRain)[:,1:]

NN = len(NAreal)-1   # number of nodes
SS = len(ICost)-1     # number of (NB)S
TTD = len(Retent[:,0])-1 # number of time periods for retention/infiltration modelling
RProfAnt=len(RProfAll[:,1]) # number of rain profiles to select from (auxiliary)

# define sets
N0=pyo.RangeSet(0,NN)    # nodes incl. sink
N=pyo.RangeSet(1,NN)    # nodes excl. sink
S=pyo.RangeSet(0,SS)    # (NB)S
T=pyo.RangeSet(0,TT)    # time periods water modelling
TE=pyo.RangeSet(1,TTE)    # time periods economics
TD=pyo.RangeSet(0,TTD)    # time periods infiltration modelling

# create sparse matrices describing the network
LShare=csc_matrix((nshare,(nrow,ncol)),shape=(NN+1,NN+1)).toarray()  # share of flow from node nn going to node n
LTime=csc_matrix((ntime,(nrow,ncol)),shape=(NN+1,NN+1)).toarray()  # time of concentration between two nodes

# select rain profile
# spread the selected RProf out over the whole modelling horizon (TT time units) and all nodes
RProf0=np.zeros((RProfAnt,1))
RProfZ=np.zeros((RProfAnt,TT-len(RProfAll[0,:])))

RProf=np.concatenate([RProf0,RProfAll,RProfZ],axis=1)[RProfsel-1,:]

# ... and span this out over all nodes where rain is considered
RNode=np.transpose(np.ones([TT+1,NN+1])*NRain)

## (we use a matrix to enable later versions with different rain profiles at different nodes)
Rain=RProf*RNode

RainNode=np.zeros([TT+1,NN+1])
for n in N:
    for t in T:
        RainNode[t,n]=NAreal[n]*Rain[n,t]

# If NShareLimit is active -> break down the given limits on area to limits on shares
if NShareLimit:
    for n in N:
        for s in S:          
            NSol[n,s]=NSol[n,s]/NAreal[n]
        NBSLimit[n]=NBSLimit[n]/NAreal[n]    
#else: ensure that we have binary values in the table (no area limits -> allow for share <=1)            
else:
    for n in N:
        for s in S:          
            NSol[n,s]=bool(NSol[n,s])
        NBSLimit[n]=bool(NBSLimit[n])    
            
#discount factor
DiscF=np.zeros(TTE+1)    
for t in TE:
    DiscF[t]=1/(1+DiscR)**(t-1)
DiscFAll=sum(DiscF) # currently, MCost is the same for all years -> we can sum up already now

if OptDischarge:
    Objective = "Optimize wrt. discharge"
else:
    Objective = "Optimize wrt. economics"


print("---------------")


# Create the model

print("Model with NBS")

if (RunSimulation!=0):
    print("Run simulation")
    shareS=np.zeros((SS+1,NN+1))
    inN=np.zeros((NN+1,TT+1))
    outN=np.zeros((NN+1,TT+1))
    inNS=np.zeros((SS+1,NN+1,TT+1))
    outNS=np.zeros((SS+1,NN+1,TT+1))
    qLoss=np.zeros((NN+1,TT+1))
    qLossS=np.zeros((SS+1,NN+1,TT+1))
    qLossAlt=np.zeros((NN+1,TT+1))
    costS=np.zeros(SS+1)
    for s in S:
        for n in N:
            shareS[s,n]=NShareFix[n,s]
    for t in T:        
        for n in N:
            inN[n,t] = Rain[n,t]*NAreal[n] + sum(LShare[nn,n]*outN[nn,t-LTime[nn,n]] for nn in N if LTime[nn,n]<=t)
            for s in S:
                inNS[s,n,t] = shareS[s,n]*inN[n,t]*(1-InfProf[s])#*NSol[n,s]
                outNS[s,n,t] = sum(inNS[s,n,t-tau]* Retent[s,tau-1] for tau in TD if tau <= t)
            outN[n,t] = sum(outNS[s,n,t] for s in S)       
        inN[0,t] = sum(LShare[nn,0]*outN[nn,t- LTime[nn,0]] for nn in N if LTime[nn,0]<=t)
        outN[0,t] = inN[0,t] 
    for t in T:
        for n in N:
            qLoss[n,t]=sum(shareS[s,n]*inN[n,t]*InfProf[s] for s in S)    
    for s in S:
        costS[s] = (ICost[s]+ MCost[s]*DiscFAll)*sum(shareS[s,n]* NAreal[n] for n in N)
   
    # read out some values for writing back to Excel
    
    costR=sum(costS[s] for s in S) 
    SinkIn=sum(inN[0,t] for t in T)
    qLossRT=sum(sum(qLoss[n,t] for t in T) for n in N) 
    
    solStatus="Simulation"
    shareSR=np.transpose(shareS)
    shareSRD=pd.DataFrame(np.row_stack((np.zeros(SS+1), shareSR)))
    
       
    qLossTimeD=pd.DataFrame(np.transpose(qLoss))
    # inflow into each node
    inNR0=np.transpose(inN)
    inNRD=pd.DataFrame(inNR0)
    
    # total rain in
    RainIn=sum(sum(Rain[n,t]*NAreal[n] for n in N) for t in T)
      
    # outflow from each node
    outNR0=np.transpose(outN)
    outNRD=pd.DataFrame(outNR0)
else:    
    print ('Run optimization.',Objective) 
    print ('Create the model')
    
    m=pyo.AbstractModel()
    
    ## variables
    m.shareS=pyo.Var(S, N, domain=pyo.NonNegativeReals) # area share of solution s in node n
    m.costN=pyo.Var(N)  # total cost in node n
    m.costS=pyo.Var(S)      # total cost of solution s
    m.inN=pyo.Var(N0, T,domain=pyo.NonNegativeReals) # flow into node n
    m.inNS=pyo.Var(S, N0, T,domain=pyo.NonNegativeReals) # flow into solution s in node n
    m.outN=pyo.Var(N0, T, domain=pyo.NonNegativeReals)   # flow out of node n
    m.outNS=pyo.Var(S, N0, T, domain=pyo.NonNegativeReals)   # flow out of solution s in node n
    m.cost=pyo.Var()        # expression for objective function - total costs
    m.qLoss=pyo.Var(N0,T)  # loss through evaporation etc. in a node
    m.disch=pyo.Var(T)  # discharge per time unit
    m.tdisch=pyo.Var()  # total discharge
    
    # define constraints and obj. functions
         
    ## constraints
    # area/share restrictions 
    # if run as simulation - fix decision variables (shares of measures in nodes) to given input values
    def c_simul_l(m,s,n):
        if s>0:    
           return m.shareS[s,n]==NShareFix[n,s]
        else:  
            return m.shareS[s,n]>=0
            
    def c_feasS(m,s,n):     # (NB)S feasible in the node? Evt. also upper limit on installed area
        return m.shareS[s,n] <= NSol[n,s]
  
    def c_feasNBS(m,n):     # limit on total area for NBS
        return sum(m.shareS[s,n] for s in S) - m.shareS[0,n] <=NBSLimit[n]
     
    def c_share(m,n):       # can't exceed available area in the node (and install "something" on all available area)
        return sum(m.shareS[s,n] for s in S) == 1
    
    def c_noNBS(m,n):   # for comparison with no-NBS-case: set share of no-NBS (= measure 0) to 1
        return m.shareS[0,n] == 1
    
    # hydrology
    def c_inN(m,n,t):    # inflow into nodes, except sink node (that has no rain)
        return m.inN[n,t] == Rain[n,t]*NAreal[n] + sum(LShare[nn,n]*m.outN[nn,t-LTime[nn,n]] for nn in N if LTime[nn,n]<=t)
    
    def c_inSink(m,t):          # inflow into sink node
        return m.inN[0,t] == sum(LShare[nn,0]*m.outN[nn,t- LTime[nn,0]] for nn in N if LTime[nn,0]<=t)
    
    def c_inNS(m,s,n,t):    # distribution of inflow to (NB)S
        return m.inNS[s,n,t] == m.shareS[s,n]*m.inN[n,t]*(1-InfProf[s])*NSol[n,s]
    
    def c_outNS(m,s,n,t):     # flow out of (NB)S
        return m.outNS[s,n,t] == sum(m.inNS[s,n,t-tau]* Retent[s,tau-1] for tau in TD if tau <= t)
    
    def c_outN(m,n,t):        # flow out of node
        return m.outN[n,t] == sum(m.outNS[s,n,t] for s in S)
    
    def c_inSinkLimit(m,t):     # limit on flow into/out of sink node
        return m.inN[0,t] <= LimitDischarge
    
    def c_outSink(m,t): # for the sake of completeness...
        return m.outN[0,t] == m.inN[0,t] 
    
    def c_disch(m,t):
        return m.disch[t]==m.outN[0,t] 
     
    def c_qLoss(m,n,t):   # calculate qLoss = total loss from evaporation etc. in a node
     #   return m.qLoss[s,n] == sum(m.shareS[s,n]*m.inN[n,t]*InfProf[s]*NSol[n,s] for t in T)
       return m.qLoss[n,t] == sum(m.shareS[s,n]*m.inN[n,t]*InfProf[s]*NSol[n,s] for s in S)
    
    # economics
    def c_costS(m,s):       # costs per solution (over all nodes)
        return m.costS[s] == (ICost[s]+ MCost[s]*DiscFAll)*sum(m.shareS[s,n]* NAreal[n] for n in N)
                              
    def c_costN(m,n):       # costs per node (over all solutions)
        return m.costN[n] == sum((( ICost[s]+ MCost[s]*DiscFAll)*m.shareS[s,n]* NAreal[n])  for s in S)
    
    def c_costLimit(m):     # limit on total costs
        return m.cost <= LimitCosts
        
    # objective functions
        
    def obj_cost(m):
        return m.cost==pyo.summation(m.costS)

    def obj_disch(m):
        return m.tdisch==pyo.summation(m.disch)
        

    ############################
    if OptDischarge:    # set limit on total costs, minimize total discharge
        m.Obj=pyo.Objective(expr=m.tdisch)
        m.C_obj=pyo.Constraint(rule=obj_disch)
        m.C_CostLimit=pyo.Constraint(rule=c_costLimit)
        m.C_Costcalc=pyo.Constraint(rule=obj_cost)
    else:       # set limit on discharge, minimize costs
        m.Obj=pyo.Objective(expr=m.cost)
        m.C_obj=pyo.Constraint(rule=obj_cost)
        m.C_SinkLimit = pyo.Constraint(T, rule = c_inSinkLimit)
    
    m.C_CostS = pyo.Constraint(S, rule = c_costS)
    m.C_CostN = pyo.Constraint(N, rule = c_costN)
    m.C_ShareSN = pyo.Constraint(N, rule = c_share)
    m.C_FeasSN = pyo.Constraint(S, N, rule = c_feasS)
    m.C_FeasNBS=pyo.Constraint(N,rule=c_feasNBS)
 
    m.C_SinkIn = pyo.Constraint(T, rule = c_inSink)
    m.C_SinkOut = pyo.Constraint(T, rule=c_outSink)
    m.C_Disch=pyo.Constraint(T,rule=c_disch)

    m.C_NodeIn = pyo.Constraint(N, T, rule = c_inN)
    m.C_NodeSolIn = pyo.Constraint(S, N, T, rule = c_inNS)
    m.C_NodeSolOut = pyo.Constraint(S, N, T, rule = c_outNS)
    m.C_NodeOut = pyo.Constraint(N, T, rule = c_outN)
    m.C_qLoss=pyo.Constraint(N,T,rule=c_qLoss)
        
    
    #### populate model with input values read in from Excel
    #### i.e., create an instance
    instance = m.create_instance()
    print('Created model instance')
        
    # write out the model to check
    if checkfiles:
        oldout=sys.stdout
        ff = open(modelfile, 'w')
        sys.stdout = ff
        #instance.C_NodeSolOut.pprint()
        instance.pprint()
        ff.close()
        sys.stdout=oldout
        print('Wrote model instance to file',modelfile)
    
    #### solve the model
    #### remember we have nonlinear constraints -> find a good/suitable NLP solver
    
    if ipopt_exe is not None:
        def optimize(instance):
            opt = pyo.SolverFactory('ipopt', solver_io='nl', executable=ipopt_exe)
            results = opt.solve(instance, tee=True)
            optimize(instance)
    else:
        opt = pyo.SolverFactory('ipopt', solver_io='nl', executable='ipopt.exe')
        results = opt.solve(instance, tee=True)
    
    if checkfiles:
        #send detailed results to file
        oldout=sys.stdout
        ff = open(resultfile, 'w')
        sys.stdout = ff
        instance.display()
        ff.close()
        sys.stdout=oldout
        print('Wrote optimization results to', resultfile)
    
    # read out values from optimization for writing back to Excel
    # each measure's share in each node (add sink (0) manually, we don't have measures there)
    shareSR=[[instance.shareS[s,n]() for s in S] for n in N]
    shareSRD=pd.DataFrame(np.row_stack((np.zeros(SS+1), shareSR)))
    
    # inflow into each node
    inNR0=[[instance.inN[n,t]() for n in N0] for t in T]
    inNRD=pd.DataFrame(inNR0)
    inNR=[[instance.inN[n,t]() for n in N] for t in T]
    inNSR=[[[instance.inNS[s,n,t]() for s in S] for n in N0] for t in T]
    
    # total rain in
    RainIn=sum(sum(Rain[n,t]*NAreal[n] for n in N) for t in T)
    
    # total water arriving at sink
    SinkIn=sum([instance.inN[0,t]() for t in T])
    
    # outflow from each node
    outNR0=[[instance.outN[n,t]() for n in N0] for t in T]
    outNRD=pd.DataFrame(outNR0)
    outNR=[[instance.outN[n,t]() for n in N] for t in T]
  #  outNSR=[[instance.outNS[s,sn,t]() for s in S] for t in T]
    
    # total loss
    qLossTime=[[instance.qLoss[n,t]() for n in N0] for t in T]
   
    #qLossR=sum(qLossTime[t] for t in T)
    qLossRT=sum(sum(instance.qLoss[n,t] () for n in N) for t in T)
    
   # qLossRTD=pd.DataFrame(qLossRT)
    qLossTimeD=pd.DataFrame(qLossTime)
    
    costR=instance.cost()
    
    solStatus =results.solver.termination_condition

    print("Solution status:",solStatus)
    
    solStatus = Objective + ' (' + solStatus + ')'

## for comparison: run with all areas fixed to measure 0 (no NBS)

print("Benchmark without NBS, run simulation")

shareS=np.zeros((SS+1,NN+1))
inN=np.zeros((NN+1,TT+1))
outN=np.zeros((NN+1,TT+1))
inNS=np.zeros((SS+1,NN+1,TT+1))
outNS=np.zeros((SS+1,NN+1,TT+1))
qLoss=np.zeros((SS+1,NN+1))
qLossTime=np.zeros((NN+1,TT+1))
costS=np.zeros(SS+1)
for n in N:
    shareS[0,n]=1
for t in T:        
    for n in N:
        inN[n,t] = Rain[n,t]*NAreal[n] + sum(LShare[nn,n]*outN[nn,t-LTime[nn,n]] for nn in N if LTime[nn,n]<=t)
        inNS[0,n,t] = inN[n,t]*(1-InfProf[0])
        outNS[0,n,t] = sum(inNS[0,n,t-tau]* Retent[0,tau-1] for tau in TD if tau <= t)
        outN[n,t] = outNS[0,n,t]       
    inN[0,t] = sum(LShare[nn,0]*outN[nn,t- LTime[nn,0]] for nn in N if LTime[nn,0]<=t)
    outN[0,t] = inN[0,t] 
    for n in N0:
        qLoss[0,n] = sum(inN[n,t]*InfProf[0] for t in T)
    for t in T:
        for n in N0:
            qLossTime[n,t] = inN[n,t]*InfProf[0]*NSol[n,0]
    costS[0] = (ICost[0]+ MCost[0]*DiscFAll)*sum(NAreal[n] for n in N)   
    ## read out some values for writing back to Excel
    costRno=costS[0] 
    SinkInno=sum(inN[0,t] for t in T)
    qLossRTno=sum(qLoss[0,n] for n in N)
    qLossRTimeno=qLossTime
    solStatusno="Simulation"
    shareSRno=np.transpose(shareS)
    shareSRDno=pd.DataFrame(np.row_stack((np.zeros(SS+1), shareSR)))
    
    # inflow into each node
    inNR0no=np.transpose(inN)          
    inNRDno=pd.DataFrame(inNR0no)  
    # outflow from each node
    outNR0no=np.transpose(outN)
    outNRDno=pd.DataFrame(outNR0no)
  
# combine some key information
objRcom=[[' ','Total costs (Euro)', 'Total rain in (l)', 'Total discharge (l)', 'Discharge/rain', 'Loss (l)', 'Status'], 
        ['With NBS', costR, RainIn, SinkIn, SinkIn/RainIn, qLossRT, solStatus],
        ['No NBS', costRno, RainIn, SinkInno, SinkInno/RainIn, qLossRTno, solStatusno]
        ]
objRcom=pd.DataFrame(objRcom)

RainNodeD=pd.DataFrame(RainNode)
### ------- end of part on no nbs

### create some graphs  
Tplot=np.array(T)
# for graph 1: total rain in and total volume arriving at sink node 
NodeInP=np.array(inNR0[:])
SinkInPlot= NodeInP[:,0] # inNR0[0]
RainPlotAll=np.array(sum(Rain[n,:]*NAreal[n] for n in N))[:]

# for graph 2: inflow vs. outflow for a given node - fixed to node 1 
NodeInPlot=NodeInP[:,1] #inNR0[sn] #
NodeOutP=np.array(outNR0[:])
NodeOutPlot=NodeOutP[:,1]
RainPlotN=np.array(Rain[1,:]*NAreal[1])[:]

# for graph 3: inflow to sink for NBS vs. non-NBS
NodeInnoP=np.transpose(np.array(inNR0no[:]))    
SinkInnoPlot=NodeInnoP[0,:]
   
# Create the figures and plot the data
# graph 1: total rain in and total volume arriving at sink node 
fig1, ax1 = plt.subplots()
plt.plot(Tplot,SinkInPlot, label='Discharge (inflow to sink)')
plt.plot(Tplot,RainPlotAll, label='Rain over whole area')

ax1.set(xlabel='Time periods', ylabel='Volume',
       title='Total rain and discharge profile')
ax1.grid()
plt.axhline(y=LimitDischarge, color='black',linestyle='--',label='Limit on discharge')
plt.figlegend(loc="center right")
fig1.savefig('RainInSink.png')
    
# graph 2: rain, inflow vs. outflow for a given node
fig2, ax2 = plt.subplots()
plt.plot(Tplot,NodeInPlot,label='Inflow')
plt.plot(Tplot,NodeOutPlot, label='Outflow')
plt.plot(Tplot, RainPlotN, label='Rain in node')
plt.figlegend(loc="center right")
titleP2=str('Rain profile and in- vs. outflow in node 1')# + str(sn))
ax2.set(xlabel='Time periods', ylabel='Volume',
       title=titleP2)
ax2.grid()
    
# for graph 3: inflow to sink for NBS vs. non-NBS
fig3, ax3 = plt.subplots()
plt.plot(Tplot,SinkInPlot,label='With NBS')
plt.plot(Tplot,SinkInnoPlot,label='Without NBS')
ax3.set(xlabel='Time periods', ylabel='Volume',
       title='Discharge (inflow to sink), NBS vs. no NBS')
ax3.grid()
plt.axhline(y=LimitDischarge, color='black',linestyle='--',label='Limit on discharge')
plt.figlegend(loc="center right")
fig3.savefig('Sinknode')

#### send output back to Excel
ExcelOut=str(Outputfile + '.xlsx')
print('Writing back to:', ExcelOut)

with pd.ExcelWriter(ExcelOut, engine = 'xlsxwriter') as writer:
    objRcom.to_excel(writer,sheet_name='General',index=False,header=False)
    shareSRD.to_excel(writer,sheet_name='Shares')#, index=False)
    inNRD.to_excel(writer,sheet_name='Nodes_inflow (l per period)')
    outNRD.to_excel(writer,sheet_name='Nodes_outflow (l per period)')#, index=False)
    RainNodeD.to_excel(writer,sheet_name='Rain in nodes (l per period)')
    qLossTimeD.to_excel(writer,sheet_name='Loss in nodes (l per period)')
    inNRDno.to_excel(writer,sheet_name='Nodes_in_noNBS (l per period)')
    outNRDno.to_excel(writer,sheet_name='Nodes_out_noNBS (l per period)')#, index=False)    
    wb=writer.book
    ws=wb.add_worksheet("Graphs")    
    ws.insert_image(0,0,'Sinknode.png')
    ws.insert_image(0,10,'RainInSink.png')

print('Done\n---------------')